/*******************************************************************************
 * Copyright 2013 Seapine Software, Inc. All rights reserved.
 * 
 * IMPORTANT:  This Software is supplied to you by Seapine Software, Inc. ("Seapine")
 * in consideration of your agreement to the following terms, and your use, installation,
 * modification or redistribution of this Software constitutes acceptance of these terms.
 * If you do not agree with these terms, please do not use, install, modify or redistribute
 * this Software.
 * 
 * In consideration of your agreement to abide by the following terms, and subject to these
 * terms, Seapine grants you a personal, non-exclusive license, under Seapine copyrights in
 * this original Seapine software (the "Software"), to use, reproduce, modify and redistribute
 * the Software, with or without modifications, in source and/or binary forms; provided that
 * if you redistribute the Software in its entirety and without modifications, you must retain
 * this notice and the following text and disclaimers in all such redistributions of the Software.
 * Neither the name, trademarks, service marks or logos of Seapine, Inc. may be used to endorse
 * or promote products derived from the Software without specific prior written permission from
 * Seapine.  Except as expressly stated in this notice, no other rights or licenses, express or
 * implied, are granted by Seapine herein, including but not limited to any patent rights that
 * may be infringed by your derivative works or by other works in which the Software may be
 * incorporated.
 * 
 * The Software is provided by Seapine on an "AS IS" basis.  SEAPINE MAKES NO WARRANTIES,
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF NON-INFRINGEMENT,
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, REGARDING THE SOFTWARE OR ITS USE AND
 * OPERATION ALONE OR IN COMBINATION WITH YOUR PRODUCTS. 
 * 
 * IN NO EVENT SHALL SEAPINE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) ARISING IN ANY WAY OUT OF THE USE,
 * REPRODUCTION, MODIFICATION AND/OR DISTRIBUTION OF THE SOFTWARE, HOWEVER CAUSED AND WHETHER
 * UNDER THEORY OF CONTRACT, TORT (INCLUDING NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN
 * IF SEAPINE HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/

#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <iostream>
#include <fstream>
#include <stack>
#include <string>
#include <vector>
using namespace std;

////////////////////////////////////////////////////////////////////////////////

class KeywordPrefixNode;

struct KeywordMapping
{
   string sStartKeyword;
   string sEndKeyword;
   string sReplacement;
};

struct EnvironmentKeywordMapping
{
   string sStartKeyword;
   string sEndKeyword;
   const char *szEnvVar;
};

struct KeywordStackEntry
{
   string *pStringReplacement;
   KeywordPrefixNode *pMatchingStartNode;
   size_t startPos;
};

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/* Begin user customizations. */

////////////////////////////////////////////////////////////////////////////////

// Example of a function that determines the value of a keyword expansion.
string getCopyright()
{
   return string("$Copyright (c) 2013 WysiCorp. All rights reserved.$");
}

// You can bind keywords to custom replacement text here. You can also embed
//  keywords inside the replacement text and they will be expanded as well.
static const KeywordMapping gCustomKeywordList[] =
{
   { "$Company", "$", "WysiCorp" },
   { "$Copyright", "$", getCopyright() },
   { "$EventTime", "$", "$EventTime: $TriggerEvent$ occurred at $TriggerTime$ on $TriggerDate$$" }
};
static const unsigned int gNumCustomKeywords = sizeof(gCustomKeywordList)
                                               / sizeof(KeywordMapping);

////////////////////////////////////////////////////////////////////////////////

// The keywords listed below refer to environment variables set when the
//  trigger fires, but you can add custom variables as well.
static const EnvironmentKeywordMapping gEnvKeywordList[] =
{
   { "$TriggerDesc", "$", "SSCM_TRIGGERDESC" },
   { "$TriggerEvent", "$", "SSCM_EVENT" },
   { "$TriggerBeforeAfter", "$", "SSCM_BEFOREAFTER" },
   { "$TriggerDate", "$", "SSCM_DATE" },
   { "$TriggerTime", "$", "SSCM_TIME" },
   { "$TriggerUser", "$", "SSCM_USER" },
   { "$TriggerClientIP", "$", "SSCM_CLIENTIP" },
   { "$TriggerClientMachine", "$", "SSCM_CLIENTMACHINE" },
   { "$TriggerMainline", "$", "SSCM_MAINLINE" },
   { "$TriggerBranchType", "$", "SSCM_BRANCHTYPE" },
   { "$TriggerBranch", "$", "SSCM_BRANCH" },
   { "$TriggerRepository", "$", "SSCM_REPOSITORY" },
   { "$TriggerFile", "$", "SSCM_FILE" }
};
static const unsigned int gNumEnvKeywords = sizeof(gEnvKeywordList)
                                            / sizeof(EnvironmentKeywordMapping);

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/* End user customizations. You should not need to alter the code below this line. */

////////////////////////////////////////////////////////////////////////////////

enum
{
   returnOk = 0,
   returnBadInputFilepath,
   returnIoError
};

////////////////////////////////////////////////////////////////////////////////

class KeywordPrefixNode
{
public:
   KeywordPrefixNode();
   virtual ~KeywordPrefixNode();
   
   void addKeyword(const string &sStartKeyword, const string &sEndKeyword, const string &sExpansion);
   KeywordPrefixNode *walk(const unsigned char letter) const;
   string *getExpansion() const { return m_pStringExpansion; }
   KeywordPrefixNode *getEndKeywordNode() const { return m_pEndKeywordNode; }
   size_t getKeywordSize() const { return m_keywordSize; }
   bool isFullKeyword() const { return m_pEndKeywordNode != 0; }
   
private:
   void removeNodes();
   KeywordPrefixNode *walkAndAddKeyword(const string &sKeyword);
   
   KeywordPrefixNode *m_pNodes[256];
   string *m_pStringExpansion;
   KeywordPrefixNode *m_pEndKeywordNode;
   size_t m_keywordSize;
};

KeywordPrefixNode::KeywordPrefixNode()
   : m_pStringExpansion(0), m_pEndKeywordNode(0), m_keywordSize(0)
{
   memset(m_pNodes, 0, sizeof(KeywordPrefixNode*) * 256);
}

KeywordPrefixNode::~KeywordPrefixNode()
{
   removeNodes();
}

void KeywordPrefixNode::removeNodes()
{
   for (int i = 0; i < 256; ++i)
   {
      if (m_pNodes[i])
      {
         delete m_pNodes[i];
         m_pNodes[i] = 0;
      }
   }
   if (m_pStringExpansion)
   {
      delete m_pStringExpansion;
      m_pStringExpansion = 0;
   }
}

void KeywordPrefixNode::addKeyword(const string &sStartKeyword,
                                   const string &sEndKeyword,
                                   const string &sExpansion)
{
   assert(sStartKeyword.length() > 0 && sEndKeyword.length() > 0);
   
   KeywordPrefixNode *pEndKeywordNode = walkAndAddKeyword(sEndKeyword);
   KeywordPrefixNode *pStartKeywordNode = walkAndAddKeyword(sStartKeyword);
   
   if (pStartKeywordNode && pEndKeywordNode)
   {
      pStartKeywordNode->m_pStringExpansion = new string(sExpansion);
      pStartKeywordNode->m_pEndKeywordNode = pEndKeywordNode;
      pStartKeywordNode->m_keywordSize = sStartKeyword.length();
      pEndKeywordNode->m_pEndKeywordNode = pEndKeywordNode;
      pEndKeywordNode->m_keywordSize = sEndKeyword.length();
   }
}

KeywordPrefixNode *KeywordPrefixNode::walk(const unsigned char letter) const
{
   assert(letter >= 0 && letter < 256);
   
   return m_pNodes[letter];
}

KeywordPrefixNode *KeywordPrefixNode::walkAndAddKeyword(const string &sKeyword)
{
   size_t length = sKeyword.length();
   KeywordPrefixNode *current = this;
   
   for (size_t i = 0; i < length; ++i)
   {
      unsigned char letter = static_cast<unsigned char>(sKeyword.at(i));
      assert(letter >= 0 && letter < 256);
      
      if (!current->m_pNodes[letter])
         current->m_pNodes[letter] = new KeywordPrefixNode();
      current = current->m_pNodes[letter];
   }
   
   return current;
}

////////////////////////////////////////////////////////////////////////////////

int doExpansions(ifstream &inStream, ofstream &outStream, KeywordPrefixNode *pKeywordTreeRoot)
{
   int returnCode = returnOk;
   
   stack<KeywordStackEntry> keywordStack;
   KeywordPrefixNode *pCurrentNode = pKeywordTreeRoot;
   KeywordPrefixNode *pPendingNode(0);
   string sIntermediate;
   size_t intermediatePos = 0;
   
   // Keep going as long as we still have more of the file to read, OR we have
   //  more keywords we can expand.
   while (outStream.good() && (inStream.good() || (!sIntermediate.empty()
      && intermediatePos < sIntermediate.length())))
   {
      bool bRewind = false;
      
      // If we need to see more of the file than we have in memory, grab
      //  another character from the input.
      if (intermediatePos >= sIntermediate.length())
      {
         char letter = inStream.get();

         if (inStream.good())
            sIntermediate += letter;
      }
      
      KeywordPrefixNode *pNextNode(0);
      
      // Scan for keywords
      if (intermediatePos < sIntermediate.length())
      {
         char letter = sIntermediate.at(intermediatePos);
         pNextNode = pCurrentNode->walk(letter);
         
         // We found something that looks like a keyword, make note of it
         if (pNextNode && pNextNode->isFullKeyword())
            pPendingNode = pNextNode;
      }
      
      // Go to next letter
      ++intermediatePos;
      
      // We found a character that is not part of a keyword, process any
      //  previous keywords.
      if (!pNextNode)
      {
         // Start back at root
         pNextNode = pKeywordTreeRoot;
         
         if (pPendingNode)
         {
            string *pStringReplacement = pPendingNode->getExpansion();
            // This character might also be part of an expansion, so make
            //  sure to rewind afterward.
            bRewind = true;
            
            // Is a starting keyword - push to stack
            if (pStringReplacement)
            {
               KeywordStackEntry entry;
               entry.pStringReplacement = pStringReplacement;
               entry.pMatchingStartNode = pPendingNode;
               
               assert(intermediatePos - 1
                        >= pPendingNode->getKeywordSize());
               
               // Here we mark where to expand the keyword from.
               entry.startPos = intermediatePos - 1
                              - pPendingNode->getKeywordSize();
               
               keywordStack.push(entry);
            }
            // Is an ending node - pop from stack
            else if (!keywordStack.empty())
            {
               KeywordStackEntry &entry = keywordStack.top();
               KeywordPrefixNode *pStartNode = entry.pMatchingStartNode;
               
               if (pStartNode
                  && pStartNode->getEndKeywordNode() == pPendingNode
                  && entry.pStringReplacement)
               {
                  size_t start = entry.startPos;
                  size_t length = intermediatePos - entry.startPos - 1;
                  assert(start + length <= sIntermediate.length());
                  
                  // Perform the actual expansion
                  sIntermediate.replace(start, length,
                                        *entry.pStringReplacement);
                  // Move intermediatePos back
                  if (entry.pStringReplacement->length() > 0)
                     intermediatePos = entry.startPos + 1;
                  else // Special case for zero-length string replacement
                     intermediatePos = entry.startPos;
                  
                  // We are manually setting the position, so don't rewind.
                  bRewind = false;
         
                  keywordStack.pop();
               }
            }
            
            pPendingNode = 0;
         }
         
         // No stack of keywords that need to be expanded, write this letter
         //  out.
         if (keywordStack.empty() && !sIntermediate.empty()
            && intermediatePos >= sIntermediate.length())
         {
            outStream << sIntermediate.at(0);
            sIntermediate.erase(0, 1);
            intermediatePos = 0;
         }
      }
      
      if (bRewind && intermediatePos > 0)
         --intermediatePos;
      
      pCurrentNode = pNextNode;
   }
   
   // If we still have intermediate input, send to output.
   outStream << sIntermediate;
   
   if (inStream.bad())
   {
      cerr << "Error: Failed to read file." << endl;
      returnCode = returnIoError;
   }
   else if (outStream.bad())
   {
      cerr << "Error: Failed to write file." << endl;
      returnCode = returnIoError;
   }
   
   return returnCode;
}

int processFile(const char *szFilepath, KeywordPrefixNode *pKeywordTreeRoot)
{
   int returnCode = returnOk;
   ifstream streamFile;
   
   streamFile.open(szFilepath, ifstream::in | ifstream::binary);
   
   // Test if this is a binary file
   while (streamFile.good())
   {
      char letter = streamFile.get();
      
      if (streamFile.good() && letter == '\0')
      {
         cerr << "Input is a binary file." << endl;
         // Not fatal, but we aren't going to modify this file, so leave as is
         //  and allow check-in.
         streamFile.clear(ios::badbit);
      }
   }
   
   if (returnCode == returnOk && !streamFile.bad())
   {
      ofstream streamOutFile;
      string sNewFilepath = szFilepath;
      sNewFilepath += ".out";
      
      streamOutFile.open(sNewFilepath.c_str(), ifstream::out | ifstream::binary);
      
      streamFile.clear();
      streamFile.seekg(0, streamFile.beg);
      
      returnCode = doExpansions(streamFile, streamOutFile, pKeywordTreeRoot);
      
      streamFile.close();
      streamOutFile.close();
      
      if (returnCode == returnOk)
      {
         // We've made it this far. Seems like we were successful, so move the
         //  output file to the original file's location.
         if (remove(szFilepath) == 0)
            rename(sNewFilepath.c_str(), szFilepath);
      }
   }
   else
      streamFile.close();
   
   return returnCode;
}

int getExpansions(KeywordPrefixNode *pInKeywordTreeRoot)
{
   int returnCode = returnOk;
   
   // Expand these keywords to their respective environment variables.
   for (unsigned int i = 0; i < gNumEnvKeywords; ++i)
   {
      string sValue;
#ifdef WIN32
      char *szGetEnv;
      size_t length;
      _dupenv_s(&szGetEnv, &length, gEnvKeywordList[i].szEnvVar);
#else
      char *szGetEnv = getenv(gEnvKeywordList[i].szEnvVar);
#endif
      
      if (szGetEnv)
         sValue = szGetEnv;

      pInKeywordTreeRoot->addKeyword(gEnvKeywordList[i].sStartKeyword,
                                      gEnvKeywordList[i].sEndKeyword,
                                      sValue);

#ifdef WIN32
      free(szGetEnv);
#endif
   }
   
   // Expand any custom keywords.
   for (unsigned int i = 0; i < gNumCustomKeywords; ++i)
   {
      pInKeywordTreeRoot->addKeyword(gCustomKeywordList[i].sStartKeyword,
                                      gCustomKeywordList[i].sEndKeyword,
                                      gCustomKeywordList[i].sReplacement);
   }
   
   return returnCode;
}

int main(int argc, char **argv)
{
   int returnCode = returnOk;
   KeywordPrefixNode *pKeywordTreeRoot = new KeywordPrefixNode();
   
   // Get our keyword expansions.
   returnCode = getExpansions(pKeywordTreeRoot);
   
   if (returnCode == returnOk)
   {
#ifdef WIN32
      char *szInputFilepath;
      size_t length;
      _dupenv_s(&szInputFilepath, &length, "SSCM_LOCALFILE");
#else
      const char *szInputFilepath = getenv("SSCM_LOCALFILE");
#endif
      
      if (szInputFilepath)
         returnCode = processFile(szInputFilepath, pKeywordTreeRoot);
      else
      {
         cerr << "Error: The SSCM_LOCALFILE environment variable is not set." << endl;
         returnCode = returnBadInputFilepath;
      }

#ifdef WIN32
      free(szInputFilepath);
#endif
   }
   
   delete pKeywordTreeRoot;
   
   return returnCode;
}
